/**
 * $Id: controller.c,v 1.14 2007/04/06 09:05:26 superkkt Exp $
 *
 * Connection Controller Thread
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "controller.h"
#include "receiver.h"
#include "support.h"
#include "pxe.h"
#include "list.h"
#include "log.h"

/* My DCE */
static PXE_DCE dce = -1;

/**
 * set_dce - My DCE 번호를 셋팅
 *
 * @parm idx: 설정할 DCE 번호
 * @return: None
 */
void
set_dce(PXE_DCE idx)
{
    dce = idx;
}

/**
 * get_dce - My DCE 번호를 확인
 *
 * @parm: None
 * @return: My DCE 번호
 */
PXE_DCE
get_dce(void)
{
    return dce;
}

/**
 * @brief 클라이언트에게 에러 메세지 전송
 * @param fd - 소켓 디스크립터
 * @param msg - 전송 할 메세지
 * @return 없음
 */
static void
write_errmsg(int fd, const char *msg)
{
    Log("%s", msg);
    write(fd, msg, strlen(msg));
    write(fd, "\r\n", 2);
}

/**
 * @brief 소켓에서 패킷 수신을 담당할  Receiver 쓰레드 실행
 * @param fd - 소켓 디스크립터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
start_receiver(int fd)
{
    int result;
    pthread_t tid;
    int *arg = NULL;

    arg = malloc(sizeof(int));
    if (!arg)
    {
	Log("Failed to allocate memory: %s", strerror(errno));
	return -1;
    }

    *arg = fd;
    result = pthread_create(&tid, NULL, receiver, (void *) arg);
    if (result)
    {
	Log("Failed to create new receiver thread: %s", strerror(result));
	if (arg)
	    free(arg);
	return -1;
    }

    return 0;
}

/**
 * master_controller - PXE Master Mode에서 Agent의 연결을 관리
 *
 * @parm arg: 
 *	ThreadArg 구조체 포인터 (extra_arg 멤버에 MasterControl 구조체 포함)
 * @return: Always NULL
 */
void *
master_controller(void *arg)
{
    char ip[16];
    char alias[256];
    char buf[BUFSIZ + 1];
    int connfd;
    int idx;
    socklen_t len;
    struct sockaddr_in cliaddr;
    int listenfd;
    ThreadArg *targ = NULL;
    MasterControl *ct = NULL;

    /* 0번이 마스터 */
    set_dce(0);	

    /* 인자 분석 */
    targ = (ThreadArg *) arg;
    ct = (MasterControl *) targ->extra_arg;

    listenfd = socket_listen_wrap(ct->port);
    if (listenfd < 0)
    {
	Log("Failed to listen socket: %s", strerror(errno));
	exit(EXIT_FAILURE);
    }

    while (1)
    {
	/* 여기서 맺어진 커넥션은 계속 유지되어야 함 */
	len = sizeof(cliaddr);
	connfd = Accept_wrap(listenfd, (struct sockaddr *) &cliaddr, &len);
	if (connfd < 0)
	{
	    Log("Failed to accept new connection: %s", strerror(errno));
	    continue;
	}

	/* 클라이언트 IP 획득 */
	memset(ip, 0, sizeof(ip));
	inet_ntop(AF_INET, &cliaddr.sin_addr, ip, sizeof(ip));

	/* Alias Name 수신 */
	memset(alias, 0, sizeof(alias));
	if (read(connfd, alias, sizeof(alias)) <= 0)
	{
	    close(connfd);
	    continue;
	}

	{
	    /*
	     * Alias Name 유효성 확인
	     */
	    char *ptr = strstr(alias, "\r\n");
	    if (!ptr)
	    {
		Log("Connection refused: Invalid alias format");
		close(connfd);
		continue;
	    }
	    *ptr = '\0';
#define MIN_ALIASNAME	3
	    if (strlen(alias) < MIN_ALIASNAME)
	    {
		snprintf(buf, sizeof(buf), "Connection refused: alias name must "
			"longer than %d", MIN_ALIASNAME);
		write_errmsg(connfd, buf);
		close(connfd);
		continue;
	    }
	}

	/* 이미 연결되어있는 에이전트이면 연결 거부 */
	if (is_connected_agent(ip, alias))
	{
	    write_errmsg(connfd, "Connection refused: already connected IP or Alias");
	    close(connfd);
	    continue;
	}

	/* 비어있는 DCE LIST 검색해서 셋팅 */
	idx = set_dce_list(alias, ip, connfd);
	if (idx < 0)
	{
	    snprintf(buf, BUFSIZ, "Failed to establish DCE connection: "
		    "reached max DCE number (%d)", MAX_DCE);
	    write_errmsg(connfd, buf);
	    close(connfd);
	    continue;
	}

	/* 할당된 DCE 번호를 클라이언트에게 통보 */
	snprintf(buf, BUFSIZ, "%d\r\n", idx);
	write(connfd, buf, strlen(buf));

	/* 연결이 맺어진 소켓마다 하나씩의 receiver를 실행시킨다 */
	if (start_receiver(connfd))
	{
	    snprintf(buf, BUFSIZ, "Failed to start receiver thread");
	    write_errmsg(connfd, buf);
	    inactivate_dlist(idx);
	    close(connfd);
	    continue;
	}

#ifdef DEBUG
	fprintf(stderr, "[DEBUG] %s connected (DCE: %d)\n", ip, idx);
#endif

	/* XXX - 위에서 맺어진 connfd는 끊으면 안됨 */
    }

    /* Never reached */
    FREE(arg);

    /* Makes compiler happy */
    return NULL;
}

/**
 * agent_controller - PXE Agent Mode에서 Master와의 연결을 관리
 *
 * @parm arg: 
 *	ThreadArg 구조체 포인터 (extra_arg 멤버에 AgentControl 구조체 포함)
 * @return: Always NULL
 */
void *
agent_controller(void *arg)
{
    int dce;
    int connfd;
    int reconnect;
    ThreadArg *targ = NULL;
    AgentControl *ct = NULL;
    char buf[BUFSIZ + 1];
    char *ptr = NULL;

    /* 인자 분석 */
    targ = (ThreadArg *) arg;
    ct = (AgentControl *) targ->extra_arg;

    reconnect = FALSE;
    while (1)
    {
	while ((connfd = Connect_wrap(ct->ip, ct->port)) < 0)
	{
	    /* 최초 접속이 실패한 경우 */
	    if (reconnect == FALSE)
	    {
		Log("Failed to connect to master server: %s", 
			strerror(errno));
		exit(EXIT_FAILURE);
	    }

	    Log("Failed to reconnect to master server: %s", strerror(errno));
	    sleep(ct->reconnect_delay);
	}

	/*
	 * Alias 전송
	 */
	snprintf(buf, sizeof(buf), "%s\r\n", ct->alias);
	if (writen_wrap(connfd, buf, strlen(buf)) != strlen(buf))
	{
	    Log("Failed to send agent's alias name: %s", strerror(errno));
	    exit(EXIT_FAILURE);
	}

	/* 할당된 DCE 번호 획득 */
	memset(buf, 0, sizeof(buf));
	if (read(connfd, buf, BUFSIZ) <= 0)
	{
	    Log("Failed to check DCE number: %s", strerror(errno));
	    exit(EXIT_FAILURE);
	}

	/* XXX -  위에서 맺어진 connfd는 끊으면 안됨 */

	ptr = strstr(buf, "\r\n");
	if (!ptr)
	{
	    Log("Failed to retrieve DCE number: Invalied protocol");
	    exit(EXIT_FAILURE);
	}
	*ptr = '\0';

	/* DCE 번호가 digit가 아니면 에러 */
	dce = (int) strtol(buf, (char **) NULL, 10);
	/* FIXME -
	 * 원래는 strtol의 리턴값이 0이고 errno가 EINVAL이면 문자열이 숫자가 아닌
	 * 경우로 판단하는데 strtol이 숫자가 아닌 문자열을 받아도 errno를 셋팅하지
	 * 않는 문제가 발견되었다. 그래서 임시방편으로 0번 DCE 번호는 마스터 전용
	 * 이기 때문에 0으로 나오는 경우는 에러로 간주하기로 했다.
	 */
	if (dce == 0)
	{
	    Log("Invalid DCE number: %s", buf);
	    exit(EXIT_FAILURE);
	}

	/* 마스터 서버 정보 기록 */
	set_master_dce(ct->ip, connfd);

	/* 할당받은 DCE 번호 기록 */
	set_dce(atoi(buf));

	/* 연결이 맺어진 소켓마다 하나씩의 receiver를 실행시킨다 */
	if (start_receiver(connfd))
	{
	    Log("Failed to start receiver thread");
	    exit(EXIT_FAILURE);
	}

#ifdef DEBUG
	fprintf(stderr, "[DEBUG] DCE number %s is assigned\n", buf);
#endif

	/* 재접속 요청이 들어오면 마스터 서버에 재접속 시도 */
	recv_signal(CT_THREAD);
	reconnect = TRUE;

#ifdef DEBUG
	fprintf(stderr, "[DEBUG] Trying to reconnect to the master..\n");
#endif
    }

    /* Never reached */
    FREE(ct->alias);
    FREE(arg);

    /* Makes compiler happy */
    return NULL;
}

/**
 * am_i_master - 호출자가 PXE Master인지 확인
 *
 * @parm: None
 * @return: 만약 Master라면 TRUE, Agent라면 FALSE
 */
    BOOL
am_i_master(void)
{
    if (get_dce() == 0)
	return TRUE;

    return FALSE;
}

/**
 * is_localhost_ip - 정수형 IP가 255.255.255.255 인지 확인
 *
 * @parm ip: 확인할 IP
 * @return: 255.255.255.255 이면 TRUE, 아니면 FALSE
 *
 * localhost를 127.0.0.1 대신 255.255.255.255를 쓰는 이유는
 * 실제로 한 서버에서 마스터, 에이전트가 모두 작동하는 경우
 * 에이전트의 주소는 127.0.0.1이 되기 때문이다.
 */
    BOOL
is_localhost_ip(const unsigned char *ip)
{
    int i;

    for (i = 0; i < 4; i++)
    {
	if (ip[i] != 255)
	    return FALSE;
    }

    return TRUE;
}

/**
 * is_master_ip - 정수형 IP가 Master(0.0.0.0) IP인지 확인
 *
 * @parm ip: 정수형 IP
 * @return: Master IP인 경우 TRUE, 아닌 경우 FALSE
 */
    BOOL
is_master_ip(const unsigned char *ip)
{
    int i;

    for (i = 0; i < 4; i++)
    {
	if (ip[i] != 0)
	    return FALSE;
    }

    return TRUE;
}
