/**
 * $Id: list.c,v 1.31 2007/04/12 00:53:37 superkkt Exp $
 *
 * DCE / THREAD List Thread
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/time.h>

#include "list.h"
#include "pxe.h"
#include "support.h"
#include "log.h"

/* 10 미만은 PXE 필수 쓰레드용으로 예약 */
#define MIN_TAR_LIST	10

/* 0은 마스터서버용 */
#define MIN_DCE_LIST	1

/* 
 * THREAD LIST
 */
static int last_tlist_idx = MIN_TAR_LIST;
static ThreadList t_list[MAX_THREAD];
static pthread_mutex_t t_list_mutex;

/*
 * DCE LIST
 */
static int last_dlist_idx = MIN_DCE_LIST;
static DCEList d_list[MAX_DCE];
static pthread_mutex_t d_list_mutex;

/**
 * list_init - THREAD / DCE LIST 초기화
 *
 * @parm: None
 * @return: 성공 시 0, 실패 시 -1
 */
int
list_init(void)
{
    int i, j;
    int result;

    /* 
     * DCE LIST 초기화 
     */
    if ((result = pthread_mutex_init(&d_list_mutex, NULL)))
    {
	Log("Failed to initialize mutex for DCE list: %s", 
		strerror(result));
	return -1;
    }

    for (i = 0; i < MAX_DCE; i++)
    {
	d_list[i].active = FALSE;
	memset(d_list[i].alias, 0, sizeof(d_list[i].alias));
	memset(d_list[i].ip, 0, sizeof(d_list[i].ip));
	d_list[i].fd = -1;
	pthread_mutex_init(&d_list[i].mutex, NULL);
    }

    /* 
     * THREAD LIST 초기화 
     */
    if ((result = pthread_mutex_init(&t_list_mutex, NULL)))
    {
	Log("Failed to initialize mutex for DCE list: %s", 
		strerror(result));
	return -1;
    }

    for (i = 0; i < MAX_THREAD; i++)
    {
	t_list[i].active = FALSE;
	t_list[i].tid = 0;
	pthread_mutex_init(&t_list[i].mutex, NULL);
	pthread_cond_init(&t_list[i].cond, NULL);
	t_list[i].sig_status = FALSE;
	t_list[i].head = t_list[i].tail = 0;

	for (j = 0; j < MAX_CIRCULAR_QUEUE; j++)
	{
	    t_list[i].buf[j] = NULL;
	    t_list[i].len[j] = 0;
	}
    }

    return 0;
}

/**
 * search_dce_idx - 해당 IP로 접속된 DCE를 검색
 *
 * @parm ip: 검색할 IP
 * @return: 
 * 	검색된 DCE가 있을 경우 해당 DCE 번호, 검색된 내역이 없을 경우 -1
 */
PXE_DCE
search_dce_idx(const char *ip)
{
    int result;
    PXE_DCE i, ret_val;

    if ((result = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    ret_val = -1;
    for (i = 1; i < MAX_DCE; i++)
    {
	if (strcmp(ip, d_list[i].ip) == 0)
	{
	    ret_val = i;
	    break;
	}
    }

    if ((result = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    return ret_val;
}

/**
 * get_dce_list - 현재 연결된 DCE IP 리스트를 추출
 *
 * @param ptr - (OUT) IP 리스트 (호출자 사용 후 FREE)
 * @return: 리스트의 개수
 */
int
get_dce_list(PXE_DCE_LIST ***ptr)
{
    int i, j;
    int cnt;
    int result;

    if ((result = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    /* 현재 연결된 에이전트 수 */
    cnt = 0;
    
    /* 0은 마스터 서버 */
    for (i = 1; i < MAX_DCE; i++)
    {
	if (d_list[i].active)
	    cnt++;
    }
    if (cnt == 0)
	goto cleanup;

    /* +1 is terminating NULL pointer entry */
    *ptr = malloc(sizeof(PXE_DCE_LIST *) * (cnt + 1));
    if (!*ptr)
	goto cleanup;
    (*ptr)[cnt] = NULL;

    /* 0은 마스터 서버 */
    for (i = 1, j = 0; i < MAX_DCE; i++)
    {
	if (d_list[i].active)
	{
	    int idx = j;

	    (*ptr)[j++] = malloc(sizeof(PXE_DCE_LIST));
	    if ((*ptr)[idx] == NULL)
		goto cleanup;
	    (*ptr)[idx]->ip = strdup(d_list[i].ip);
	    (*ptr)[idx]->alias = strdup(d_list[i].alias);
	}
    }

cleanup:
    if ((result = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    return cnt;
}

/**
 * free_dce_list - get_dce_list에 의해 생성된 IP 리스트를 해제
 *
 * @parm list: IP 리스트 포인터
 * @return: 없음
 */
void
free_dce_list(PXE_DCE_LIST **list)
{
    int i;

    if (list)
    {
	for (i = 0; list[i] != NULL; i++)
	{
	    if (list[i]->ip)
		free(list[i]->ip);
	    if (list[i]->alias)
		free(list[i]->alias);
	    free(list[i]);
	}

	free(list);
    }
}

/**
 * _is_tlist_active - 해당 쓰레드가 Active 상태인지 확인
 *
 * @parm idx: 체크할 쓰레드의 번호
 * @return:
 * 	해당 쓰레드가 Active 상태이면 TRUE, Inavtive 상태이면 FALSE
 *
 * is_tlist_active()의 sub-routine으로서 MUTEX로 임계영역 설정을
 * 하는 루틴이 없다. 모듈의 내부에서 호출되는 경우에 MUTEX 설정이
 * 중복되지 않도록 하기 위해 사용된다.
 */
static BOOL
_is_tlist_active(PXEThread idx)
{
    if (t_list[idx].active)
	return TRUE;

    return FALSE;
}

/**
 * is_tlist_active - THREAD LIST에서 해당 쓰레드가 Active 상태인지 확인
 *
 * @parm idx: 체크할 쓰레드의 번호
 * @return:
 * 	해당 쓰레드가 Active 상태이면 TRUE, Inavtive 상태이면 FALSE
 */
BOOL
is_tlist_active(PXEThread idx)
{
    int status;
    BOOL result;

    if ((status = pthread_mutex_lock(&t_list_mutex)))
    {
	Log("Failed to lock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    result = _is_tlist_active(idx);

    if ((status = pthread_mutex_unlock(&t_list_mutex)))
    {
	Log("Failed to unlock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return result;
}

/**
 * activate_tlist - THREAD LIST에서 해당 쓰레드를 Active로 전환
 *
 * @parm idx: Active 시킬 쓰레드의 번호
 * @return: None
 */
static void
activate_tlist(PXEThread idx)
{
    t_list[idx].active = TRUE;
}


/**
 * inactivate_tlist - THREAD LIST에서 해당 쓰레드를 Inactive로 전환
 *
 * @parm idx: Inactive 시킬 쓰레드의 번호
 * @return: None
 */
void
inactivate_tlist(PXEThread idx)
{
    int i;
    int status;

    if ((status = pthread_mutex_lock(&t_list_mutex)))
    {
	Log("Failed to lock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    t_list[idx].active = FALSE;
    t_list[idx].sig_status = FALSE;
    t_list[idx].head = t_list[idx].tail = 0;

    for (i = 0; i < MAX_CIRCULAR_QUEUE; i++)
    {
	FREE(t_list[idx].buf[i]);
	t_list[idx].len[i] = 0;
    }
    if ((status = pthread_mutex_unlock(&t_list_mutex)))
    {
	Log("Failed to unlock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}


/**
 * select_tlist_index - THREAD LIST Index 번호 선택
 *
 * @parm idx: 사용할 Index 번호
 * @return: 
 *	성공 시 선택된 Index 번호, 
 *	실패(리스트가 FULL이거나 지정한 Index가 이미 사용중인 경우) 시 -1
 *
 * idx가 -1이면 자동으로 비어있는 리스트를 찾아서 사용하고
 * 0 이상이면 지정된 값을 인덱스 번호로 사용
 */
static PXEThread
select_tlist_index(PXEThread idx)
{
    PXEThread i, index;

    index = -1;

    if (last_tlist_idx == MAX_THREAD)
	last_tlist_idx = MIN_TAR_LIST;

    /* 인덱스 번호를 지정하지 않은 경우 자동으로 검색 */
    if (idx < 0)
    {
	for (i = last_tlist_idx++; i < MAX_THREAD; i++)
	{
	    if (!_is_tlist_active(i))
	    {
		index = i;
		break;
	    }
	}

	/* 예를들어 last_tlist_idx가 100이면 100부터 MAX까지 리스트를 검색한다.
	 * 만약 100부터 MAX 사이에 사용가능한 값이 없다면 MIN에서 100 사이에
	 * 사용가능한 값이 있는지 다시 체크한다.
	 */
	if (i == MAX_THREAD)
	{
	    for (i = MIN_TAR_LIST; i < last_tlist_idx; i++)
	    {
		if (!_is_tlist_active(i))
		{
		    index = i;
		    break;
		}
	    }
	}
    }
    /* 지정한 경우 해당 값을 인덱스 번호로 사용 */
    else
    {
	if (_is_tlist_active(idx))
	    return -1;

	index = idx;
    }

    return index;
}

/**
 * make_thread_argument - 쓰레드 인자 생성
 *
 * @parm arg: User Argument
 * @parm len: @parm arg 사이즈
 * @parm idx: 해당 쓰레드의 THREAD LIST Index 번호
 * @return: 성공 시 생성된 쓰레드인자 포인터, 실패 시 NULL 포인터
 */
static ThreadArg *
make_thread_argument(void *arg, size_t len, PXEThread idx)
{
    ThreadArg *ptr = NULL;
    size_t arg_len;

    if (arg && len > 0)
	/* -1 is dummy character for extra agrument */
	arg_len = sizeof(ThreadArg) - 1 + len;
    else
	arg_len = sizeof(ThreadArg);

    ptr = calloc(1, arg_len);
    if (!ptr)
	return NULL;

    ptr->idx = idx;
    if (arg && len > 0)
	memcpy(ptr->extra_arg, arg, len);

    return ptr;
}

/**
 * set_tlist_tid - THREAD LIST에 pthread tid를 셋팅
 *
 * @parm idx: THREAD LIST Index
 * @parm tid: pthread tid
 * @return: None
 *
 * 모듈 내부에서만 사용하는 함수라서 MUTEX 설정 루틴이 없다.
 * 따라서 호출자가 tlist MUTEX가 설정되었는지 확인 후 호출해야 함
 */
static void
set_tlist_tid(PXEThread idx, pthread_t tid)
{
    t_list[idx].tid = tid;
}

/**
 * get_tlist_tid - 해당 THREAD LIST의 pthread tid를 반환
 *
 * @parm idx: THREAD LIST Index
 * @return: pthread tid
 */
pthread_t 
get_tlist_tid(PXEThread idx)
{
    int status;
    pthread_t tid;

    if ((status = pthread_mutex_lock(&t_list_mutex)))
    {
	Log("Failed to lock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    tid = t_list[idx].tid;

    if ((status = pthread_mutex_unlock(&t_list_mutex)))
    {
	Log("Failed to unlock mutex for Thread list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return tid;
}

/** 
 * set_thread_list - 
 * 	비어있는 THREAD LIST 검색해서 빈 공간이 있을 경우 Active로 전환
 *	Active로 전환 후 사용자가 지정한 쓰레드 생성
 * 
 * @parm idx: 
 *	-1로 지정하면 자동으로 THREAD LIST의 빈공간을 찾아서 사용
 *	만약 0 이상의 값을 지정하면 그 값을 THREAD LIST의 Index로 사용
 * @parm function: 쓰레드 함수
 * @parm arg: 쓰레드 함수에 전달될 인자
 * @parm len: @parm arg의 길이
 * @parm options: 
 *	쓰레드 실행시 사용할 옵션 (OR 연산)
 *	    * DETACH_THREAD: 쓰레드 생성 후 DETACH
 * @return: 성공 시 셋팅한 THREAD LIST Index, 실패 시 -1
 *
 * 이 함수를 통해 생성된 쓰레드는 인자로 받은 포인터를 사용 후 
 * 직접 메모리를 해제해야만 한다.
 */
int
set_thread_list(PXEThread idx, void *(*function)(void *), 
	void *arg, size_t len, int options)
{
    PXEThread index;
    pthread_t tid;
    int result;
    ThreadArg *ptr = NULL;

    if ((result = pthread_mutex_lock(&t_list_mutex)))
    {
	Log("Failed to lock mutex for Thread list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    /* THREAD LIST Index 선택 */
    index = select_tlist_index(idx);
    if (index < 0)
    {
	Log("Failed to create new thread: no free space in thread list");
	if ((result = pthread_mutex_unlock(&t_list_mutex)))
	{
	    Log("Failed to unlock mutex for Thread list: %s", 
		    strerror(result));
	    exit(EXIT_FAILURE);
	}
	return -1;
    }

    /* 선택된 LIST 활성화 */
    activate_tlist(index);

    /* Thread Argument 작성 */
    ptr = make_thread_argument(arg, len, index);
    if (!ptr)
    {
	Log("Failed to create new thread: %s", strerror(errno));
	if ((result = pthread_mutex_unlock(&t_list_mutex)))
	{
	    Log("Failed to unlock mutex for Thread list: %s", 
		    strerror(result));
	    exit(EXIT_FAILURE);
	}
	return -1;
    }

    /* 쓰레드 생성 */
    result = pthread_create(&tid, NULL, function, (void *) ptr);
    if (result)
    {
	inactivate_tlist(index);
	Log("Failed to create new thread: %s", strerror(result));
	if ((result = pthread_mutex_unlock(&t_list_mutex)))
	{
	    Log("Failed to unlock mutex for Thread list: %s", 
		    strerror(result));
	    exit(EXIT_FAILURE);
	}
	return -1;
    }

    /* Thread ID 셋팅 */
    set_tlist_tid(index, tid);

    /* DETACH 옵션이 있는 경우 쓰레드 DETACH */
    if (options & PXE_DETACH_THREAD)
	(void) pthread_detach(tid);

    if ((result = pthread_mutex_unlock(&t_list_mutex)))
    {
	Log("Failed to unlock mutex for Thread list: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Thread ID %d created successfully\n", index);
#endif

    return index;
}

/**
 * _is_dlist_active - DCE LIST에서 해당 DCE가 활성인지 확인
 *
 * @parm idx: 체크 할 DCE 번호
 * @return: 해당 DCE가 활성일 경우 TRUE, 비활성일 경우 FALSE
 */
static BOOL
_is_dlist_active(PXE_DCE idx)
{
    if (d_list[idx].active)
	return TRUE;

    return FALSE;
}

/**
 * is_dlist_active - DCE LIST에서 해당 DCE가 활성화되었는지 확인
 * 
 * @parm idx: 확인할 DCE 번호
 * @return: 해당 DCE가 활성화되어있는 경우 TRUE, 비활성인 경우 FALSE
 */
BOOL
is_dlist_active(PXE_DCE idx)
{
    int status;
    BOOL result;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    result = _is_dlist_active(idx);

    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return result;
}


/**
 * activate_dlist - DCE LIST에서 해당 DCE를 활성화
 *
 * @parm idx: 활성화시킬 DCE 번호
 * @param alias - 활성하시 DCE LIST에 등록할 Alias Name
 * @parm ip: 활성화시 DCE LIST에 등록할 IP
 * @parm fd: 활성화시 DCE LIST에 등록할 Socket Descriptor
 * @return: None
 */
static void
activate_dlist(PXE_DCE idx, const char *alias, const char *ip, int fd)
{
    d_list[idx].active = TRUE;
    if (alias)
	snprintf(d_list[idx].alias, sizeof(d_list[idx].alias), "%s", alias);
    strncpy(d_list[idx].ip, ip, strlen(ip));
    d_list[idx].fd = fd;
}

/**
 * @brief DCE 리스트의 소켓용 뮤텍스 잠금
 * @param idx - DCE 리스트 인덱스 번호
 * @return 없음
 */
void
dlist_mutex_lock(PXE_DCE idx)
{
    int status;

    if ((status = pthread_mutex_lock(&d_list[idx].mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * @brief DCE 리스트의 소켓용 뮤텍스 해제
 * @param idx - DCE 리스트 인덱스 번호
 * @return 없음
 */
void
dlist_mutex_unlock(PXE_DCE idx)
{
    int status;

    if ((status = pthread_mutex_unlock(&d_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * set_dce_list - 비어있는 DCE LIST를 찾아서 활성화
 *
 * @param alias - DCE LIST에 등록할 Alias Name
 * @parm ip: 활성화시 DCE LIST에 등록할 IP
 * @parm fd: 활성화시 DCE LIST에 등록할 Socket Descriptor
 * @return: 성공시 할당받은 DCE LIST Index, 실패시 -1
 */
int
set_dce_list(const char *alias, const char *ip, int fd)
{
    int i;
    int ret_val;
    int status;

    ret_val = -1;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    if (last_dlist_idx == MAX_DCE)
	last_dlist_idx = MIN_DCE_LIST;

    /* 0번은 마스터 */
    for (i = last_dlist_idx++; i < MAX_DCE; i++)
    {
	if (!_is_dlist_active(i))
	{
	    activate_dlist(i, alias, ip, fd);
	    ret_val = i;
	    goto cleanup;
	}
    }

cleanup:
    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return ret_val;
}

/**
 * set_master_dce - DCE LIST에 Master DCE 정보를 기록
 *
 * @parm ip: Master의 IP
 * @parm fd: Master와 연결된 Socket Descriptor
 * @return: None
 */
void
set_master_dce(const char *ip, int fd)
{
    int status;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    activate_dlist(0, NULL, ip, fd);

    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * get_dlist_fd - 해당 DCE의 정보에서 Socket Desciptor를 반환
 *
 * @parm idx: Socket Descriptor를 가져올 DCE 번호
 * @return: Socket Descriptor
 */
int
get_dlist_fd(PXE_DCE idx)
{
    int fd;
    int status;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    fd = d_list[idx].fd;
    
    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return fd;
}

/**
 * get_dlist_ip - 해당 DCE의 정보에서 IP를 반환
 *
 * @parm idx: IP를 가져올 DCE 번호
 * @return: IP (호출자 사용 후 free)
 */
char *
get_dlist_ip(PXE_DCE idx)
{
    char *ip = NULL;
    int status;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    ip = strdup(d_list[idx].ip);

    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return ip;
}

/**
 * @brief 이미 연결된 에이전트 IP 또는 Alias Name인지 확인
 * @param ip - 확인할 IP
 * @param alias - 확인할 Alias Name
 * @return
 *  연결된 에이전트인 경우 TRUE,\n
 *  연결되지 않은 에이전트인 경우 FALSE
 */
BOOL
is_connected_agent(const char *ip, const char *alias)
{
    int i;
    int ret_val;
    int status;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    ret_val = FALSE;
    for (i = MIN_DCE_LIST; i < MAX_DCE; i++)
    {
	if (d_list[i].active == FALSE)
	    continue;
	if (strcmp(d_list[i].ip, ip) == 0 ||
		strcmp(d_list[i].alias, alias) == 0)
	{
	    ret_val = TRUE;
	    break;
	}
    }

    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    return ret_val;
}

/**
 * inactivate_dlist - DCE LIST에서 해당 DCE를 비활성화
 *
 * @parm idx: 비활성화 시킬 DCE 번호
 * @return: None
 */
void
inactivate_dlist(PXE_DCE idx)
{
    int status;

    if ((status = pthread_mutex_lock(&d_list_mutex)))
    {
	Log("Failed to lock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    d_list[idx].fd = -1;
    d_list[idx].active = FALSE;
    memset(d_list[idx].alias, 0, sizeof(d_list[idx].alias));
    memset(d_list[idx].ip, 0, sizeof(d_list[idx].ip));

    if ((status = pthread_mutex_unlock(&d_list_mutex)))
    {
	Log("Failed to unlock mutex for DCE list: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * send_signal - 쓰레드에 신호 전달
 *
 * @parm idx: 신호를 전달할 쓰레드의 THREAD LIST Index
 * @return: None
 *
 * 이 함수는 단순히 신호만 전달할 뿐 어떠한 데이터도 전달하지 않음
 */
void
send_signal(PXEThread idx)
{
    int status;

    if ((status = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    t_list[idx].sig_status = TRUE;
    (void) pthread_cond_signal(&t_list[idx].cond);

    if ((status = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * recv_signal - 쓰레드로 오는 신호를 수신
 *
 * @parm idx: 신호를 수힌할 쓰레드의 THREAD LIST Index
 * @return: None
 *
 * 이 함수는 단순히 신호만 수신할 뿐 어떠한 데이터도 수신하지 않음
 * 또한 신호를 받을때까지 블럭됨
 */
void
recv_signal(PXEThread idx)
{
    int status;

    if ((status = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }

    while (t_list[idx].sig_status == FALSE)
	pthread_cond_wait(&t_list[idx].cond, &t_list[idx].mutex);
    t_list[idx].sig_status = FALSE;

    if ((status = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(status));
	exit(EXIT_FAILURE);
    }
}

/**
 * _push_queue - 원형큐에 패킷 삽입
 *
 * @parm idx: 패킷을 삽입할 쓰레드의 THREAD LIST Index
 * @parm buf: 패킷 데이터
 * @parm len: 패킷 사이즈
 * @return: 성공 시 0, 실패(큐 오버플로우) 시 -1
 */
static int
_push_queue(PXEThread idx, void *buf, PXELength len)
{
    int head;
    int next_head;

    head = t_list[idx].head;
    next_head = head + 1;

    /* 원형큐의 끝에 도달했으면 다시 처음으로 이동 */
    if (next_head == MAX_CIRCULAR_QUEUE)
	next_head = 0;

    /* 큐가 오버플로우 되는 경우 */
    if (next_head == t_list[idx].tail)
	return -1;

    /* Entire Packet */
    t_list[idx].buf[head] = buf;

    /* PXE Header + Data Size */
    t_list[idx].len[head] = len;

    /* Overflow 상태가 아님으로 Queue Index 증가 */
    t_list[idx].head = next_head;

    return 0;
}

/**
 * push_queue - 패킷을 대상 쓰레드의 원형큐에 삽입
 *
 * @parm idx: 패킷을 삽입할 쓰레드의 THREAD LIST Index
 * @parm buf: 패킷 데이터
 * @return: 성공 시 0, 실패(큐 오버플로우) 시 -1
 */
int
push_queue(PXEThread idx, void *buf)
{
    PXEPacket *pxe = NULL;
    int status;
    int result;
    PXELength len;

    pxe = (PXEPacket *) buf;
    len = sizeof(PXEPacket) + pxe->length;

    if ((result = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    status = _push_queue(idx, buf, len);
    if (status == 0)
	(void) pthread_cond_signal(&t_list[idx].cond);

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Push packet(%d bytes) "
	    "to thread %d's queue\n", len, idx);
#endif

    if ((result = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    return status;
}

/**
 * _pop_queue - 원형큐에서 패킷을 꺼내옴
 *
 * @parm idx: 패킷을 꺼내올 쓰래드의 THREAD LIST Index
 * @parm packet (out):
 *      패킷 데이터를 저장할 포인터 (호출자가 사용후 free해야 함)
 * @parm length (out): 패킷 사이즈
 * @return: 성공 시 0, 실패(데이터가 없는 경우) 시 -1
 */
static int
_pop_queue(PXEThread idx, void **packet, PXELength *len)
{
    int head;
    int tail;

    head = t_list[idx].head;
    tail = t_list[idx].tail;

    /* No data in queue */
    if (head == tail)
	return -1;

    *packet = t_list[idx].buf[tail];
    *len = t_list[idx].len[tail];

    t_list[idx].buf[tail] = NULL;
    t_list[idx].len[tail] = 0;

    /* 원형큐의 끝에 도달하면 다시 처음으로 돌아감 */
    if (++t_list[idx].tail == MAX_CIRCULAR_QUEUE)
	t_list[idx].tail = 0;

    return 0;
}

/**
 * pop_queue - 패킷을 원형큐에서 꺼내옴
 *
 * @parm idx: 패킷을 꺼내올 쓰래드의 THREAD LIST Index
 * @parm packet (out):
 *      패킷 데이터를 저장할 포인터 (호출자가 사용후 free해야 함)
 * @parm len (out): 패킷 사이즈
 */
void
pop_queue(PXEThread idx, void **packet, PXELength *len)
{
    int result;

    if ((result = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    while (_pop_queue(idx, packet, len) < 0)
	pthread_cond_wait(&t_list[idx].cond, &t_list[idx].mutex);

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Pop packet(%d bytes) "
	    "from thread %d's queue\n", *len, idx); 
#endif

    if ((result = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }
}

/**
 * pop_queue_non_block - 패킷을 원형큐에서 꺼내옴 (non-block)
 *
 * @parm idx: 패킷을 꺼내올 쓰래드의 THREAD LIST Index
 * @parm packet (out):
 *      패킷 데이터를 저장할 포인터 (호출자가 사용후 free해야 함)
 * @parm len (out): 패킷 사이즈
 * @return: 성공 시 0, 실패 시 -1
 */
int
pop_queue_non_block(PXEThread idx, void **packet, PXELength *len)
{
    int result;
    int status;

    if ((result = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    status = _pop_queue(idx, packet, len);

#ifdef DEBUG
    if (status == 0)
	fprintf(stderr, "[DEBUG] Pop packet(%d bytes) "
		"from thread %d's queue\n", *len, idx); 
#endif

    if ((result = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    return status;
}

/**
 * pop_queue_timedwait - 패킷을 원형큐에서 꺼내옴 (timedwait)
 *
 * @parm idx: 패킷을 꺼내올 쓰래드의 THREAD LIST Index
 * @parm packet (out):
 *      패킷 데이터를 저장할 포인터 (호출자가 사용후 free해야 함)
 * @parm len (out): 패킷 사이즈
 * @param sec: TIMEOUT 시간 (초)
 * @return: 성공 시 0, 타임 아웃 시 -1
 */
int
pop_queue_timedwait(PXEThread idx, void **packet, PXELength *len, int sec)
{
    int result;
    int ret_val;
    struct timespec timeout;
    struct timeval curtime;

    /* for cleanup routine */
    ret_val = -1;

    if ((result = pthread_mutex_lock(&t_list[idx].mutex)))
    {
	Log("Failed to lock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    (void) gettimeofday(&curtime, NULL);
    timeout.tv_sec = curtime.tv_sec + sec;
    timeout.tv_nsec = curtime.tv_usec * 1000;

    while (_pop_queue(idx, packet, len) < 0)
    {
	result = pthread_cond_timedwait(&t_list[idx].cond, &t_list[idx].mutex, 
		&timeout);
	if (result == ETIMEDOUT)
	   goto cleanup; 
    }

    /* success */
    ret_val = 0;

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Pop packet(%d bytes) "
	    "from thread %d's queue\n", *len, idx); 
#endif

cleanup:
    if ((result = pthread_mutex_unlock(&t_list[idx].mutex)))
    {
	Log("Failed to unlock mutex for Thread list member: %s", 
		strerror(result));
	exit(EXIT_FAILURE);
    }

    return ret_val;
}
