/**
 * $Id: pxe.c,v 1.28 2007/04/06 09:05:26 superkkt Exp $
 *
 * PXE Main
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "pxe.h"
#include "list.h"
#include "sender.h"
#include "receiver.h"
#include "controller.h"
#include "log.h"

typedef struct
{
    PXEThread idx;              /* Index */
    void *(*function)(void *);  /* Thread Function */
} EssentialThread;

/* 필수 쓰레드 구조체 */
static EssentialThread thread_list[] = 
{
    {0, NULL}	/* Dummy */
};

/**
 * signal_register - 유닉스 시그널 핸들러 등록
 *
 * @parm: None
 * @return: 성공 시 0, 실패 시 -1
 */
static int
signal_register(void)
{
    struct sigaction sa;

    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;

    /* SIGPIPE 신호 무시 */
    sa.sa_handler = SIG_IGN;
    if(sigaction(SIGPIPE, &sa, NULL) == -1) 
	return -1;

    return 0;
}

/**
 * pxe_init - PXE 초기화
 *
 * @parm log: 로그 기록 함수 포인터
 * @return: 성공 시 0, 실패시 -1
 */
int 
pxe_init(LogFunction log)
{
    int result;
    PXEThread i;

    if (signal_register())
	return -1;

    if (list_init())
	return -1;

    set_log_function(log);

    for (i = 0; i < ARRAY_SIZE(thread_list); i++)
    {
	/* 필수 쓰레드 실행
	 *
	 * Sender와 Reciver가 각각 하나씩만 존재할때는 이들이 필수 쓰레드로 
	 * 실행이 되었었다. 하지만 지금은 I/O마다 하나씩 존재하도록 변경
	 * 되었기 때문에 필수 쓰레드에서 제거되었다. 그로 인해서 현재는 필수
	 * 쓰레드가 하나도 없지만 나중을 위해 이 루틴을 남겨두었다. 
	 */
	if (thread_list[i].function)
	{
	    result = set_thread_list(thread_list[i].idx, 
		    thread_list[i].function, NULL, 0, 0);
	    if (result < 0)
		return -1;
	}
    }

    return 0;
}

/**
 * pxe_master - PXE 마스터 구동
 * 
 * @parm port: PXE 서버가 사용할 포트
 * @return: 성공 시 0, 실패 시 -1
 */
int 
pxe_master(int port)
{
    int result;
    MasterControl arg;

    arg.port = port;

    result = set_thread_list(CT_THREAD, master_controller, 
	    &arg, sizeof(MasterControl), 0);

    /* FIXME -
     * 여기서 생성된 controller 쓰레드가 늦게 실행이 되어서 이 함수의
     * 호출자가 먼저 실행이 되는 경우 마스터 - 에이전트 연결이 안된것처럼
     * 보일 수 있다. 그래서 임시방편으로 usleep.
     */
    usleep(1000);

    return result < 0 ? -1 : 0;
}

/**
 * pxe_agent - PXE 에이전트 구동
 *
 * @param alias - 에이전트 Alias Name
 * @parm ip: PXE 서버 IP
 * @parm port: PXE 서버 PORT
 * @parm delay: 접속이 끊어졌을때 재접속을 시도할 주기(초)
 * @return: 성공 시 0, 실패 시 -1
 */
int 
pxe_agent(const char *alias, const char *ip, int port, int delay)
{
    int result;
    AgentControl arg;

    memset(arg.ip, 0, sizeof(arg.ip));
    strncpy(arg.ip, ip, sizeof(arg.ip));
    arg.alias = strdup(alias);
    arg.port = port;
    arg.reconnect_delay = delay;

    result = set_thread_list(CT_THREAD, agent_controller, 
	    &arg, sizeof(AgentControl), 0);

    /* FIXME -
     * 여기서 생성된 controller 쓰레드가 늦게 실행이 되어서 이 함수의
     * 호출자가 먼저 실행이 되는 경우 마스터 - 에이전트 연결이 안된것처럼
     * 보일 수 있다. 그래서 임시방편으로 usleep.
     */
    usleep(1000);

    return result < 0 ? -1 : 0;
}

/** 
 * pxe_load_thread - 지정된 함수를 PXE의 쓰레드로 로딩
 *
 * @parm idx: 
 *	THREAD LIST Index
 *	* 0 이상의 숫자를 지정할 경우 해당값을 THREAD LIST Index로 사용
 *	* -1을 지정할 경우 자동으로 비어있는 Index를 찾아서 사용
 * @parm function: 쓰레드 함수 포인터
 * @parm arg: 쓰레드 함수에 전달할 인자
 * @parm len: @parm arg의 사이즈
 * @return: 성공 시 THREAD LIST Index, 실패 시 -1
 *
 * 이 함수를 통해 생성된 쓰레드는 인자로 받은 포인터를 사용 후 
 * 직접 메모리를 해제해야만 한다.
 */
int 
pxe_load_thread(PXEThread idx, void *(*function)(void *), 
	void *arg, size_t len, int options)
{
    /* 자동으로 빈 리스트를 찾아서 실행 */
    return set_thread_list(idx, function, arg, len, options);
}

/**
 * pxe_get_argument - 
 *	쓰레드함수 인자에서 해당 쓰레드의 User Argument를 추출
 *
 * @parm arg: 쓰레드함수가 받은 인자
 * @return: 해당 쓰레드의 User Argument 포인터
 */
void *
pxe_get_argument(PXEArg *arg)
{
    return arg->extra_arg;
}

/**
 * pxe_get_src_thread - PXE 핸들러에서 Source Thread 정보를 리턴
 *
 * @parm pxe: PXE 핸들러 포인터
 * @return: PXE Source Thread
 */
PXEThread
pxe_get_src_thread(PXE *pxe)
{
    return pxe->peer_thread;
}

/**
 * pxe_get_src_ip - PXE 핸들러에서 Source IP 정보를 문자열로 추출
 *
 * @parm pxe: PXE 핸들러 포인터
 * @parm ip: 추출한 문자열 IP를 저장할 메모리 포인터
 * @parm len: @parm ip의 길이
 * @return: 없음
 */
void
pxe_get_src_ip(PXE *pxe, char *ip, size_t len)
{
   get_ip(ip, len, pxe->peer_ip); 
}

/**
 * pxe_open - PXE read/write를 위한 handler 생성
 *
 * @parm arg: 쓰레드함수가 받은 인자 (해당 쓰레드의 THREAD LIST Index 참조)
 * @return: 성공 시 PXE Handler 포인터, 실패 시 NULL 포인터
 */
PXE *
pxe_open(PXEArg *arg)
{
    PXE *pxe = NULL;

    pxe = malloc(sizeof(PXE));
    if (!pxe)
	return NULL;

    /* Destination 정보 초기화 */
    pxe->peer_dce = 0;
    pxe->peer_thread = 0;

    /* Caller의 THREAD LIST Index를 기록 */
    pxe->my_thread = arg->idx;

    /* 옵션 초기화 */
    pxe->options = 0;

    return pxe;
}

/**
 * pxe_get_options - PXE 옵션 플래그 리턴
 *
 * @parm pxe: PXE 핸들러 포인터
 * @return: PXE 옵션 플래그
 */
int
pxe_get_options(PXE *pxe)
{
    return pxe->options;
}

/**
 * pxe_set_options - PXE 옵션 플래그 셋팅
 *
 * @parm pxe: PXE 핸들러 포인터
 * @parm options: 옵션 플래그
 * @return: 없음
 */
void 
pxe_set_options(PXE *pxe, int options)
{
    pxe->options = options;
}

/**
 * pxe_set_destination - PXE write를 위해 대상 호스트의 정보를 셋팅
 *
 * @parm pxe: PXE Handler
 * @parm ip: 대상 호스트의 IP
 * @parm dst_thread: 대상 THREAD LIST Index
 * @parm err(OUT): 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return: 성공 시 0, 실패(연결되지 않은 IP) 시 -1
 */
int
pxe_set_destination(PXE *pxe, const char *ip, PXEThread dst_thread)
{
    int dce;
    unsigned char in_ip[4];

    /* 문자열 IP -> 정수형 IP */
    set_ip(in_ip, ip);

    if (is_localhost_ip(in_ip))
	dce = get_dce();	/* to myself */
    else if (is_master_ip(in_ip))
	dce = 0;	/* to master */
    else
    {
	if (am_i_master())
	{
	    dce = search_dce_idx(ip);
	    if (dce < 0)
	    {
		errno = ENOTCONN;
		return -1;
	    }
	}
	else
	    dce = 0;	/* to master for routing to other agent */
    }

    memcpy(pxe->peer_ip, in_ip, sizeof(pxe->peer_ip));
    pxe->peer_dce = dce;
    pxe->peer_thread = dst_thread;

    return 0;
}

/**
 * make_pxe_packet - PXE Packet 생성
 *
 * @parm pxe: PXE Handler
 * @parm buf: 전송할 데이터
 * @parm len: @parm buf의 사이즈
 * @return: 성공 시 생성된 PXE Packet의 포인터, 실패 시 NULL Pointer
 */
static void *
make_pxe_packet(PXE *pxe, const void *buf, size_t len)
{
    size_t total_len;
    PXEPacket *packet = NULL;
    void *ptr = NULL;
    void *data = NULL;

    /* PXE Packet + User Data */
    total_len = sizeof(PXEPacket) + len;
    ptr = calloc(1, total_len);
    if (!ptr)
	return NULL;

    packet = (PXEPacket *) ptr;
    data = (char *) ptr + sizeof(PXEPacket);

    packet->emblem = PXE_EMBLEM;
    packet->id = DATA_ID;
    memcpy(packet->dst_ip, pxe->peer_ip, sizeof(packet->dst_ip));
    packet->src_dce = get_dce();
    packet->dst_dce = pxe->peer_dce;
    packet->src_thread = pxe->my_thread;
    packet->dst_thread = pxe->peer_thread;

    /* 데이터 길이 */
    packet->length = len;

    if (buf && len > 0)
	memcpy(data, buf, len);

    return ptr;
}

/**
 * pxe_write - PXE 패킷 전송
 *
 * @parm pxe: PXE Handler
 * @parm buf: 전송할 데이터
 * @parm len: @parm buf의 사이즈
 * @return: 성공 시 0, 실패 시 -1
 *
 * 연결되지 않은 DCE에 write를 시도하면 ENOTCONN errno를 셋팅
 */
int
pxe_write(PXE *pxe, const void *buf, size_t len)
{
    int result;
    int total_len;
    void *packet = NULL;

    packet = make_pxe_packet(pxe, buf, len);
    if (!packet)
	return -1;
    total_len = sizeof(PXEPacket) + len;

    /* 대상 호스트와 연결되어 있는지 확인 */
    if (get_dce() != pxe->peer_dce) 
    {
	if (!is_dlist_active(pxe->peer_dce))
	{
	    FREE(packet);
	    errno = ENOTCONN;
	    return -1;
	}
    }

    result = send_packet(packet, total_len);
    FREE(packet);
    return result;
}

/**
 * pxe_read - PXE 패킷 수신
 *
 * @parm pxe: PXE Handle
 * @parm ptr(out): 사용자 데이터 포인터
 * @return: 
 *	수신한 패킷에서 사용자 데이터의 사이즈 (PXE 헤더 제외), 
 *	NON_BLOCK 사용 시 수신 패킷이 없는 경우 0,
 *	TIMEDWAIT 사용 시 타임아웃되는 경우 -1
 *
 * NON_BLOCK 옵션이 사용되지 않은 경우 데이터를 모두 수신할때까지 블럭된다.
 * 수신에 성공한 경우 PXE 헤더를 건너뛰고 실제 데이터의 위치를 가르키는 메모리 
 * 포인터를 @parm ptr에 저장한다.
 *
 * 사용자 데이터가 있는 경우에는 사용 후 반드시 pxe_free()를 통해 해제를 
 * 시켜줘야 한다.
 */
PXELength
pxe_read(PXE *pxe, void **ptr)
{
    PXELength len;
    PXELength data_len;
    void *packet = NULL;
    PXEPacket *header = NULL;

    /* 
     * 패킷 수신 
     */
    if (pxe->options & PXE_NON_BLOCK)
    {
	/* 큐에 수신된 패킷이 없음 */
	if (pop_queue_non_block(pxe->my_thread, &packet, &len))
	    return 0;
    }
    else
    {
	pop_queue(pxe->my_thread, &packet, &len);
	if (!packet)
	    return 0;
    }

    /* 수신한 패킷의 헤더정보를 pxe 핸들에 기록 */
    header = (PXEPacket *) packet;
    memcpy(pxe->peer_ip, header->src_ip, sizeof(pxe->peer_ip));
    pxe->peer_dce = header->src_dce;
    pxe->peer_thread = header->src_thread;

    data_len = header->length;
    if (data_len > 0)
	*ptr = (char *) packet + sizeof(PXEPacket);
    else
    {
	/* 사용자 데이터가 없기 때문에 헤더 패킷 메모리 해제 */
	FREE(packet);
	*ptr = NULL;
    }

    return data_len; 
}

/**
 * pxe_read_timedwait - PXE 패킷 수신
 *
 * @parm pxe: PXE Handle
 * @parm ptr(out): 사용자 데이터 포인터
 * @param timeout: TIMEOUT 시간 (초)
 * @return: 
 *	수신한 패킷에서 사용자 데이터의 사이즈 (PXE 헤더 제외), 
 *	TIMEDWAIT 사용 시 타임아웃되는 경우 -1
 *
 * timeout에 지정된 시간동안 데이터를 수신할때까지 블럭된다.
 * 수신에 성공한 경우 PXE 헤더를 건너뛰고 실제 데이터의 위치를
 * 가르키는 메모리 포인터를 @parm ptr에 저장한다.
 *
 * 사용자 데이터가 있는 경우에는 사용 후 반드시 pxe_free()를 
 * 통해 해제를 시켜줘야 한다.
 */
PXELength
pxe_read_timedwait(PXE *pxe, void **ptr, int timeout)
{
    PXELength len;
    PXELength data_len;
    void *packet = NULL;
    PXEPacket *header = NULL;

    if (pop_queue_timedwait(pxe->my_thread, &packet, &len, timeout) == -1)
	return -1;

    /* 수신한 패킷의 헤더정보를 pxe 핸들에 기록 */
    header = (PXEPacket *) packet;
    memcpy(pxe->peer_ip, header->src_ip, sizeof(pxe->peer_ip));
    pxe->peer_dce = header->src_dce;
    pxe->peer_thread = header->src_thread;

    data_len = header->length;
    if (data_len > 0)
	*ptr = (char *) packet + sizeof(PXEPacket);
    else
    {
	/* 사용자 데이터가 없기 때문에 헤더 패킷 메모리 해제 */
	FREE(packet);
	*ptr = NULL;
    }

    return data_len; 
}

/**
 * pxe_is_from_external - 패킷이 외부에서 수신되었는지 확인
 *
 * @parm pxe: PXE 핸들러 포인터
 * @return: 외부에서 수신되었으면 0, 내부에서 수신되었으면 -1
 */
int
pxe_is_from_external(PXE *pxe)
{
    if (get_dce() == pxe->peer_dce)
	return -1;

    return 0;
}

/**
 * pxe_free - 수신한 PXE 패킷 메모리 해제
 *
 * @parm ptr: 수신한 데이터
 * @return: None
 */
void
pxe_free(void *ptr)
{
    void *p = NULL;

    if (ptr)
    {
	/* 패킷 헤더를 포함하는 실제 할당된 메모리의
	 * 시작 위치로 이동하기 위해 헤더사이즈만큼 뺀다
	 */
	p = (char *) ptr - sizeof(PXEPacket);
	FREE(p);
    }
}

/**
 * pxe_close - PXE Handler 제거
 *
 * @parm pxe: PXE Handler
 * @return: None
 */
void
pxe_close(PXE *pxe)
{
    inactivate_tlist(pxe->my_thread);
    FREE(pxe);
}

/**
 * pxe_sleep - 필수 쓰레드가 종료할때까지 대기 후 종료
 *
 * @parm: None
 * @return: None
 *
 * 필수 쓰레드는 절대로 종료되지 않는다. 따라서 필수 쓰레드와 같이
 * 호출자를 무한정 잠들게 하는 효과를 가진다.
 */
void
pxe_sleep(void)
{
    pthread_t tid;

    tid = get_tlist_tid(CT_THREAD);
    pthread_join(tid, NULL);
}

/**
 * pxe_get_dce_list - 현재 연결된 DCE IP 리스트를 추출
 *
 * @param list - (OUT) IP 리스트 (호출자 사용 후 FREE)
 * @return: 리스트의 개수
 */
int
pxe_get_dce_list(PXE_DCE_LIST ***list)
{
    return get_dce_list(list);
}

/**
 * pxe_free_dce_list - pxe_get_dce_list에 의해 생성된 IP 리스트를 해제
 *
 * @parm list: IP 리스트 포인터
 * @return: 없음
 */
void
pxe_free_dce_list(PXE_DCE_LIST **list)
{
    free_dce_list(list);
}
