/**
 * $Id: receiver.c,v 1.27 2007/04/06 09:05:26 superkkt Exp $
 *
 * Receiver Thread
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>

#include "receiver.h"
#include "pxe.h"
#include "list.h"
#include "sender.h"
#include "controller.h"
#include "log.h"

/* 소켓 연결 유지를 위해 PING 패킷을 보낼 주기 (초) */
#define PING_PERIOD    60

/**
 * wait_packet - 연결된 DCE connection에서 패킷 수신 대기
 *
 * @param fd: 소켓 디스크립터
 * @return: 성공 시 0, 실패(Timeout or Error) 시 -1
 */
static int
wait_packet(int fd)
{
    int result;
    fd_set rset;
    struct timeval tm;

    FD_ZERO(&rset);
    FD_SET(fd, &rset);

    /*
     * 소켓이 PING_PERIOD 시간동안 idle 상태이면 select 중지 
     */
    tm.tv_sec = PING_PERIOD;
    tm.tv_usec = 0;
    result = select(fd + 1, &rset, NULL, NULL, &tm);
    if (result <= 0)
    {
	if (result == 0 || errno == EINTR)
	    /* Timeout or Interrupt */
	    return -1; 
	else
	{
	    /* Error */
	    Log("select() function failed: %s", strerror(errno));
	    return -1;
	}
    }

    return 0;
}

/**
 * @brief 스트림 형태의 PXE 헤더 정보를 PXE 구조체로 변환
 * @param pxe - PXE 구조체 포인터
 * @param header - 스트림 형태의 PXE 헤더
 * @return 없음
 */
static void
make_pxe_struct(PXEPacket *pxe, const void *header)
{
    uint32_t *emblem = NULL;
    uint32_t *id = NULL;
    uint8_t *dst_ip = NULL;
    uint8_t *src_ip = NULL;
    int16_t *dst_dce = NULL;
    int16_t *src_dce = NULL;
    int16_t *dst_thread = NULL;
    int16_t *src_thread = NULL;
    int32_t *length = NULL;

    emblem = (uint32_t *) header;
    id = (uint32_t *) ((char *) emblem + 4);
    dst_ip = (uint8_t *) ((char *) id + 4);
    src_ip = (uint8_t *) ((char *) dst_ip + 4);
    dst_dce = (int16_t *) ((char *) src_ip + 4);
    src_dce = (int16_t *) ((char *) dst_dce + 2);
    dst_thread = (int16_t *) ((char *) src_dce + 2);
    src_thread = (int16_t *) ((char *) dst_thread + 2);
    length = (int32_t *) ((char *) src_thread + 2);

    pxe->emblem = *emblem;
    pxe->id = *id;
    memcpy(pxe->dst_ip, dst_ip, 4);
    memcpy(pxe->src_ip, src_ip, 4);
    pxe->dst_dce = *dst_dce;
    pxe->src_dce = *src_dce;
    pxe->dst_thread = *dst_thread;
    pxe->src_thread = *src_thread;
    pxe->length = *length;
}

/**
 * recv_packet - 소켓에서 패킷 READ
 *
 * @parm idx: DCE LIST Index
 * @parm fd: 소켓 디스크립터
 * @parm buf (out): 패킷을 저장할 메모리 포인터 (호출자가 free해야 함)
 * return: 성공 시 수신한 패킷 바이트 수, EOF 수신 시 0, 실패시 -1
 */
static int
recv_packet(PXE_DCE idx, int fd, unsigned char **buf)
{
    PXELength len;
    PXELength size;
    void *header = NULL;
    PXEPacket pxe;

    /* 헤더를 읽어들일 버퍼 */
    header = calloc(1, PXE_PACKET_SIZE);
    if (!header)
    {
	Log("Failed to allocate memory: %s", strerror(errno));
	exit(EXIT_FAILURE);
    }

    /* PXE header만 read */
    len = readn_wrap(fd, header, PXE_PACKET_SIZE);
    if (len <= 0)
    {
	FREE(header);
	return len;
    }

    /* 이기종간의 연결을 위해 스트림 형태로 전송되는 PXE 헤더를 구조체로 변환 */
    make_pxe_struct(&pxe, header);

    /* 바이트오더 변환 */	
    pxe.emblem = ntohl(pxe.emblem);
    pxe.id = ID_NTOH(pxe.id);
    pxe.dst_dce = DCE_NTOH(pxe.dst_dce);
    pxe.src_dce = DCE_NTOH(pxe.src_dce);
    pxe.dst_thread = THREAD_NTOH(pxe.dst_thread);
    pxe.src_thread = THREAD_NTOH(pxe.src_thread);
    pxe.length = LENGTH_NTOH(pxe.length);

    /* 송신 호스트 IP 기록 */
    if (am_i_master())
    {
	char *ip = get_dlist_ip(idx);
	set_ip(pxe.src_ip, ip);
	if (ip)
	    free(ip);
    }

    /* 헤더 구조체 + 데이터 메모리 할당 */
    size = sizeof(PXEPacket) + pxe.length;
    *buf = malloc(size);
    if (!*buf)
    {
	Log("Failed to allocate memory: %s", strerror(errno));
	exit(EXIT_FAILURE);	
    }

    /* PXE header 뒤에 data가 있으면 data read */
    if (pxe.length > 0)
    {
	len = readn_wrap(fd, *buf + sizeof(PXEPacket), pxe.length);
	if (len <= 0)
	{
	    FREE(*buf);
	    FREE(header);
	    return len;
	}
    }

    /* 이미 읽어들인 header를 복사 */
    memcpy(*buf, &pxe, sizeof(PXEPacket));
#ifdef DEBUG
    {
	char dst_ip[16];
	char src_ip[16];

	get_ip(dst_ip, sizeof(dst_ip), pxe.dst_ip);
	get_ip(src_ip, sizeof(src_ip), pxe.src_ip);

	fprintf(stderr, "[DEBUG] Read %d bytes from DCE %d (src_ip: %s, "
		"dst_ip: %s, dst_dce: %d, src_thread: %d, dst_thread: %d)\n", 
		size, pxe.src_dce, src_ip, dst_ip, pxe.dst_dce, 
		pxe.src_thread, pxe.dst_thread);
    }
#endif

    FREE(header);
    return size;
}

/**
 * routing_packet - 다른 에이전트로 전달되는 패킷인지 확인
 *
 * @parm packet: 수신된 패킷
 * @return: 
 *	Master Mode이고 패킷의 타겟이 다른 에이전트인 경우 SD_THREAD 번호,
 * 	그렇지 않은 경우 패킷의 타겟 쓰레드 번호
 */
static PXEThread
routing_packet(void *packet)
{
    char ip[16];
    PXEPacket *pxe = NULL;

    pxe = (PXEPacket *) packet;

    /* 
     * Master Mode인 경우 Destination IP 분석해서 타겟이 Master가 
     * 아니면 해당 IP로 다시 패킷을 전달하도록 대상 DCE, THREAD 정보 변경
     */
    if (am_i_master() && !is_master_ip(pxe->dst_ip))
    {
	/* 문자열 IP 추출 */
	get_ip(ip, sizeof(ip), pxe->dst_ip);

	pxe->src_dce = get_dce();
	pxe->dst_dce = search_dce_idx(ip);

#ifdef DEBUG
	fprintf(stderr, "[DEBUG] Packet routed to DCE %d\n", pxe->dst_dce);
#endif

	return SD_THREAD;
    }

    return pxe->dst_thread;
}

/**
 * @brief 패킷이 PING 패킷인지 여부 확인
 * @param packet - 확인할 패킷
 * @return
 *  PING 패킷인 경우 TRUE,\n
 *  아닌 경우 FALSE
 */
static BOOL
is_ping_packet(const void *packet)
{
    PXEPacket *ptr = (PXEPacket *) packet;

    if (ptr->id == PING_ID)
	return TRUE;

    return FALSE;
}

/**
 * @brief 유효한 PXE 패킷인지 확인
 * @param packet - 확인할 패킷
 * @return
 *  유효한 PXE 패킷인 경우 TRUE,\n
 *  유효하지 않은 PXE 패킷인 경우 FALSE
 */
static BOOL
is_valid_packet(const void *packet)
{
    PXEPacket *ptr = (PXEPacket *) packet;

    if (ptr->emblem != PXE_EMBLEM)
	return FALSE;
    if (ptr->dst_thread < 0 || ptr->dst_thread > MAX_THREAD)
	return FALSE;

    return TRUE;
}

/**
 * get_packet - 
 * 	패킷을 수신한 소켓으로부터 패킷 READ해서 대상 THREAD QUEUE에 삽입
 *
 * @param fd - 소켓 디스크립터
 * @return: 정상 패킷 수신 시 0, 소켓이 끊어진 경우 -1
 */
static int 
get_packet(int fd)
{
    int i, idx = 0;
    int cnt;
    int result;
    unsigned char *buf = NULL;
    PXEThread tid;

    for (i = 0; i < MAX_DCE; i++)
    {
	if (!is_dlist_active(i))
	    continue;
	if (fd == get_dlist_fd(i))
	{
	    idx = i;
	    break;
	}
    }
    if (i == MAX_DCE)
	return -1;

    /* 패킷 수신 */
    result = recv_packet(idx, fd, &buf);
    if (result <= 0)
    {
	if (result == 0)
	    /* 연결이 끊어진 경우 */ 
	    Log("Connection closed by peer host");
	else
	    /* 에러 발생 */
	    Log("Socket error: %s", strerror(errno));

	/* DCE LIST에 비활성으로 등록 */
	close(fd);
	inactivate_dlist(idx);

	/* 마스터가 연결이 끊어진 경우 재접속 */
	if (idx == 0)
	    send_signal(CT_THREAD);

	return -1;
    }

    if (is_ping_packet(buf))
    {
	/* 
	 * PING 패킷인 경우 패킷을 버린 후 아무 작업도 하지 않음 
	 */
	FREE(buf);
#ifdef DEBUG
	fprintf(stderr, "[DEBUG] PING packet received\n");
#endif
	return 0;
    }

    if (is_valid_packet(buf) == FALSE)
    {
	FREE(buf);
	/* 
	 * 잘못된 패킷을 수신한 경우 소켓이 끊어진걸로 간주 
	 */
	close(fd);
	inactivate_dlist(idx);
	/* 마스터가 연결이 끊어진 경우 재접속 */
	if (idx == 0)
	    send_signal(CT_THREAD);
	Log("Invalid PXE Packet received, so connection closed by master");
	return -1;
    }

    tid = routing_packet(buf);
    /* 다른 호스트로 전송되는 패킷 */
    if (tid == SD_THREAD)
    {
	send_packet(buf, result);
	FREE(buf);
	return 0;
    }

    /* 대상 쓰레드가 비활성 */
    if (!is_tlist_active(tid))
    {
#ifdef DEBUG
	fprintf(stderr, "[DEBUG] Failed to rotue packet to "
		"thread %d: This thread is not activated\n", tid);
#endif
	FREE(buf);
	return 0;
    }

    /*
     * 수신한 패킷을 타겟 쓰레드의 큐에 삽입
     * 오버플로우 발생시 반복 실행
     */
    cnt = 0;
    while (push_queue(tid, buf) < 0)
    {
#ifdef DEBUG
	cnt++;
#endif
	usleep(OVERFLOW_DELAY);
    }
#ifdef DEBUG
    if (cnt > 0)
	fprintf(stderr, "[DEBUG] Thread circular queue "
		"is overflowed %d times\n", cnt);
#endif

    return 0;
}

/**
 * receiver - PXE Packet Receiver
 *
 * @parm arg: 사용 안함
 * @return: Always NULL Pointer
 */
void *
receiver(void *arg)
{
    int fd;
    int result;

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Receiver thread starting\n");
#endif

    fd = *(int *) arg; 
    while (1)
    {
	/* 패킷 수신대기 */
	result = wait_packet(fd);
	if (result < 0)
	{
	    if (am_i_master())
	    {
		send_ping(fd);
#ifdef DEBUG
		fprintf(stderr, "[DEBUG] Sent ping packet to %d FD\n", fd);
#endif
	    }
	    continue;
	}

	/* 패킷 수신 */
	if (get_packet(fd))
	    /* 소켓이 끊어진 경우 receiver 종료 */
	    break;
    }

    FREE(arg);
#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Receiver thread finished\n");
#endif
    /* Makes compiler happy */
    return NULL;
}
