/**
 * $Id: sender.c,v 1.18 2007/04/06 09:05:26 superkkt Exp $
 *
 * Sender Thread
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <arpa/inet.h>

#include "sender.h"
#include "pxe.h"
#include "list.h"
#include "controller.h"
#include "log.h"

/**
 * local_send - 로컬호스트로 패킷 전송
 *
 * @parm packet: PXE Packet
 * @return: None
 */
static void
local_send(void *packet, PXELength len)
{
    ssize_t cnt;
    char ip[16];
    void *dup_packet = NULL;
    PXEPacket *pxe = NULL;

    /*
     * sender 모듈은 패킷을 외부로 전송하고 해당 패킷의 메모리를
     * 해제한다. 그러나 로컬호스트로 전송되는 패킷은 해제되면 안되기
     * 때문에 새로 메모리를 할당해서 전달한다.
     */
    dup_packet = malloc(len);
    if (!dup_packet)
    {
	Log("Failed to allocate memory: %s", strerror(errno));
	exit(EXIT_FAILURE);
    }
    memcpy(dup_packet, packet, len);

    /* 
     * receiver가 패킷을 수신할때 리턴패킷을 위해 dst_ip를 src_ip로 바꿔준다.
     * 하지만 로컬호스트로 전송되는 패킷은 receiver를 거치지 않고 바로 쓰레드의
     * 원형큐에 삽입되기 때문에 여기서 수동으로 dst_ip -> src_ip로 전환해야 한다.
     */
    pxe = (PXEPacket *) dup_packet;
    get_ip(ip, sizeof(ip), pxe->dst_ip);
    set_ip(pxe->src_ip, ip);

    /*
     * 대상 쓰레드의 원형큐에 패킷 삽입
     */
    cnt = 0;
    while (push_queue(pxe->dst_thread, dup_packet) < 0)
    {
#ifdef DEBUG
	cnt++;
#endif
	usleep(OVERFLOW_DELAY);
    }
#ifdef DEBUG
    if (cnt > 0)
	fprintf(stderr, "[DEBUG] Thread circular queue "
		"is overflowed %d times\n", cnt);

    fprintf(stderr, "[DEBUG] Sent packet to "
	    "localhost's thread %d\n", pxe->dst_thread);
#endif
}

/**
 * @brief PXE 헤더와 데이터를 바이트의 연속인 스트림 형태로 변환
 * @param stream - (OUT) 생성된 스트림 포인터 (호출자 사용 후 FREE)
 * @param pxe - PXE 패킷 헤더
 * @param data - 실제 데이터
 * @param len - \a data 의 사이즈
 * @return
 *  성공 시 생성된 스트림의 전체 사이즈,\n
 *  실패 시 -1
 */
static int 
make_stream(void **stream, PXEPacket *pxe, const void *data, size_t len)
{
    uint32_t *emblem = NULL;
    uint32_t *id = NULL;
    uint8_t *dst_ip = NULL;
    uint8_t *src_ip = NULL;
    int16_t *dst_dce = NULL;
    int16_t *src_dce = NULL;
    int16_t *dst_thread = NULL;
    int16_t *src_thread = NULL;
    int32_t *length = NULL;
    void *packet = NULL;

    *stream = malloc(PXE_PACKET_SIZE + len);
    if (!*stream)
	return -1;

    emblem = (uint32_t *) *stream;
    id = (uint32_t *) ((char *) emblem + 4);
    dst_ip = (uint8_t *) ((char *) id + 4);
    src_ip = (uint8_t *) ((char *) dst_ip + 4);
    dst_dce = (int16_t *) ((char *) src_ip + 4);
    src_dce = (int16_t *) ((char *) dst_dce + 2);
    dst_thread = (int16_t *) ((char *) src_dce + 2);
    src_thread = (int16_t *) ((char *) dst_thread + 2);
    length = (int32_t *) ((char *) src_thread + 2);
    packet = (char *) length + 4;

    *emblem = pxe->emblem;
    *id = pxe->id;
    memcpy(dst_ip, pxe->dst_ip, 4); 
    memcpy(src_ip, pxe->src_ip, 4); 
    *dst_dce = pxe->dst_dce;
    *src_dce = pxe->src_dce;
    *dst_thread = pxe->dst_thread;
    *src_thread = pxe->src_thread;
    *length = pxe->length;
    if (data && len > 0)
	memcpy(packet, data, len);

    return PXE_PACKET_SIZE + len;
}

/**
 * send_packet - 대상 호스트로 패킷 전송
 *
 * @parm packet: 전송할 패킷
 * @parm len: 패킷 사이즈
 * @return: 성공 시 0, 실패 시 -1
 */
int
send_packet(void *packet, PXELength len)
{
    int fd;
    int ret_val;
    PXEPacket *pxe = NULL;
    PXE_DCE dst_dce;
    ssize_t cnt;
    int stream_len;
    void *stream = NULL;

    /* for cleanup routine */
    ret_val = -1;

    /* 대상 호스트의 식별코드 확인 */
    pxe = (PXEPacket *) packet;
    dst_dce = pxe->dst_dce;

    dlist_mutex_lock(dst_dce);

    /* 로컬호스트로 보내는 패킷이거나 MASTER -> MASTER인 경우 */
    if (is_localhost_ip(pxe->dst_ip) ||
	    (am_i_master() && is_master_ip(pxe->dst_ip)))
    {
	local_send(packet, len);
	ret_val = 0;
	goto cleanup;
    }

#ifdef DEBUG
    fprintf(stderr, "[DEBUG] Sending packet to DCE %d\n", dst_dce);
#endif

    /* 바이트오더 변환 */
    pxe->emblem = htonl(pxe->emblem);
    pxe->id = ID_HTON(pxe->id);
    pxe->dst_dce = DCE_HTON(pxe->dst_dce);
    pxe->src_dce = DCE_HTON(pxe->src_dce);
    pxe->dst_thread = THREAD_HTON(pxe->dst_thread);
    pxe->src_thread = THREAD_HTON(pxe->src_thread);
    pxe->length = LENGTH_HTON(pxe->length);

    /* Socket Descriptor 확인 */
    fd = get_dlist_fd(dst_dce);

    /* 이기종간의 연결을 위해 PXE 헤더와 데이터를 스트림으로 변경 */
    stream_len = make_stream(&stream, pxe, (char *) packet + sizeof(PXEPacket), 
	    LENGTH_NTOH(pxe->length));
    if (stream_len == -1)
    {
	Log("Failed to make stream: %s", strerror(errno));
	goto cleanup;
    }

    /* 패킷 전송 */
    cnt = writen_wrap(fd, stream, stream_len);
    if (cnt == -1)
    {
	/* 에러 발생이기 때문에 해당 소켓 비활성 */
	inactivate_dlist(dst_dce);
	Log("Failed to write packet: %s", strerror(errno));
	goto cleanup;
    }

#ifdef DEBUG
    {
	char dst_ip[16];
	char src_ip[16];

	get_ip(dst_ip, sizeof(dst_ip), pxe->dst_ip);
	get_ip(src_ip, sizeof(src_ip), pxe->src_ip);

	fprintf(stderr, "[DEBUG] Sent packet(%d bytes) to DCE %d (src_ip: %s, "
		"dst_ip: %s, src_dce: %d, src_thread: %d, dst_thread: %d)\n", 
		cnt, dst_dce, src_ip, dst_ip, DCE_NTOH(pxe->src_dce), 
		THREAD_NTOH(pxe->src_thread), THREAD_NTOH(pxe->dst_thread));
    }
#endif
    /* Success */
    ret_val = 0;

cleanup:
    free(stream);
    dlist_mutex_unlock(dst_dce);
    return ret_val;
}

/**
 * @brief 소켓으로 PING 패킷 전송
 * @param fd - 소켓 디스크립터
 * @return 없음
 */
void
send_ping(int fd)
{
    int i, idx = 0;
    PXEPacket packet;
    int stream_len;
    void *stream = NULL;

    for (i = 0; i < MAX_DCE; i++)
    {
	if (!is_dlist_active(i))
	    continue;
	if (fd == get_dlist_fd(i))
	{
	    idx = i;
	    break;
	}
    }
    if (i == MAX_DCE)
	return;

    dlist_mutex_lock(idx);
    memset(&packet, 0, sizeof(PXEPacket));
    packet.id = ID_HTON(PING_ID);
    stream_len = make_stream(&stream, &packet, NULL, 0); 
    if (stream_len != -1)
    {
	(void) writen_wrap(fd, stream, stream_len);
	free(stream);
    }
    dlist_mutex_unlock(idx);
}
