/**
 * $Id: support.c,v 1.12 2007/04/06 09:05:26 superkkt Exp $
 *
 * Support Functions
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdarg.h>

#include "support.h"
#include "log.h"

/**
 * Accept_wrap - accept() wrapping 함수
 *
 * @param fd: Socket descriptor
 * @param sa: 소켓 구조체
 * @param salenptr: @param sa의 length
 * @return: 실패 시 -1, 성공 시 socket descriptor
 */
int
Accept_wrap(int fd, struct sockaddr *sa, socklen_t *salenptr)
{
    int n;

again:
    if((n = accept(fd, sa, salenptr)) < 0) {
#ifdef  EPROTO
	if(errno == EPROTO || 
		errno == ECONNABORTED || 
		errno == EINTR) {
#else
	if(errno == ECONNABORTED || 
		errno == EINTR) {
#endif
	    goto again;
	}
	else {
	    return -1;
        }
    }   

    return n;
}

/**
 * Connect_wrap - 소켓 생성하고 지정한 호스트로 접속 
 *
 * @param host: 호스트네임 또는 IP
 * @param port: 접속 포트
 * @return: 실패 시 -1, 성공 시 socket descriptor
 */
int 
Connect_wrap(const char *host, int port)
{
#ifdef HAVE_GETADDRINFO
    int	sockfd, n;
    int result;
    char port_str[6];
    struct addrinfo hints, *res, *ressave;

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    snprintf(port_str, sizeof(port_str), "%d", port);
    if((n = getaddrinfo(host, port_str, &hints, &res)) != 0) {
	return -1;
    }
    ressave = res;

    do {
	sockfd = socket(res->ai_family, 
		res->ai_socktype, 
		res->ai_protocol);
	/* ignore error */
	if(sockfd < 0)
	    continue;	

	result = connect(sockfd, res->ai_addr, res->ai_addrlen);
	if(result == 0)
	    break;		
	if(result < 0 && errno == EINTR) {
	    close(sockfd);
	    freeaddrinfo(ressave);
	    return -1;
	}

	close(sockfd);		
    } while((res = res->ai_next) != NULL);

    if(res == NULL) {
	freeaddrinfo(ressave);
	return -1;
    }

    freeaddrinfo(ressave);

    return sockfd;

#else

    int fd;
    struct sockaddr_in servaddr;
    struct hostent *hptr = NULL;

    /* gethostbyname set h_errno if it failed */
    hptr = gethostbyname(host);
    if (!hptr)
	return -1;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0)
	return -1;

    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    memcpy(&servaddr.sin_addr, *hptr->h_addr_list, sizeof(servaddr.sin_addr));

    if (connect(fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
    {
	close(fd);
	return -1;
    }

    return fd;
#endif
}

/**
 * socket_listen_wrap - 특정 포트로 소켓생성 후 리슨
 *
 * @param port: 리슨할 포트 번호
 * @return: 실패 시 -1, 성공 시 socket descriptor
 */
int
socket_listen_wrap(int port)
{
    int fd;
    int ret_val;
    struct sockaddr_in servaddr;
    const int on = 1;
#ifdef TCP_KEEPALIVE
    const int keep_alive = 60;
#endif

    ret_val = -1;
    if((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return -1;

    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(port);

    if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) 
	goto cleanup;
    if(setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on)) < 0) 
	goto cleanup;
#ifdef TCP_KEEPALIVE
    if(setsockopt(fd, IPPROTO_TCP, TCP_KEEPALIVE, &keep_alive, sizeof(keep_alive)) < 0) 
	goto cleanup;
#endif

    if(bind(fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
	goto cleanup;
    if(listen(fd, 1024) < 0)
	goto cleanup;

    /* Success */
    ret_val = fd;

cleanup:
    if (ret_val == -1)
	close(fd);

    return ret_val;
}

/**
 * readn_wrap - 지정된 사이즈만큼 소켓에서 READ
 *
 * @param fd: Socket descriptor
 * @param vptr: 수신한 데이터를 저장할 버퍼
 * @param n: 수신할 데이터의 사이즈
 * @return: 
 *	실패 시 -1, 
 * 	성공 시 수신한 데이터 사이즈
 */
ssize_t
readn_wrap(int fd, void *vptr, size_t n)
{
    size_t nleft;
    ssize_t nread;
    char *ptr = NULL;

    ptr = vptr;
    nleft = n;
    while(nleft > 0) {
	if((nread = read(fd, ptr, nleft)) < 0) {
	    if(errno == EINTR) {
		nread = 0;
	    }
	    else {
		return -1;
	    }
	}
	else if(nread == 0) {
	    break;
	}

	nleft -= nread;
	ptr += nread;
    }

    return (n - nleft);
}

/**
 * writen_wrap - 지정된 사이즈만큼 소켓에 WRITE
 *
 * @param connfd: Socket descriptor
 * @param buf: write할 데이터
 * @param len: @param buf의 길이
 * @return: 
 *	실패 시 -1, 
 * 	성공 시 총 write한 바이트수 (== @parm len)
 */
ssize_t
writen_wrap(int connfd, const void *buf, size_t len)
{
    const char *ptr = buf;
    int nwrite, twrite = 0, olen = len;

    while(len > 0) {
again:
	if((nwrite = write(connfd, ptr, len)) < 0) {
	    if(errno == EINTR) {
		goto again;
	    }
	    else {
		return -1;
	    }
	}

	len -= nwrite;
	ptr += nwrite;
	twrite += nwrite;
    }

    return (twrite == olen) ? twrite : -1;
}

/**
 * set_ip - 문자열로 된 IP를 정수형으로 기록
 *
 * @parm buf: IP를 기록할 버퍼
 * @parm ip: 기록할 IP
 * @return: 성공 시 0, 실패 시 -1 
 */
int
set_ip(unsigned char *buf, const char *ip)
{
    int i;
    char *dup_ip = NULL;
    char *offset = NULL;
    char *ptr = NULL;

    dup_ip = strdup(ip);
    if (!dup_ip)
    {
	Log("Failed to allocate memory: %s", strerror(errno));
	exit(EXIT_FAILURE);
    }

    offset = dup_ip;
    for (i = 0; i < 3; i++)
    {
	ptr = strchr(offset, '.');
	if (!ptr)
	{
	    Log("Failed to set IP to PXE packet: Invalid IP Address");
	    free(dup_ip);
	    return -1;
	}
	*ptr++ = '\0';

	buf[i] = atoi(offset);
	offset = ptr;
    }
    buf[3] = atoi(offset);

    free(dup_ip);

    return 0;
}

/**
 * get_ip - 정수형으로 된 IP를 문자열로 변환
 *
 * @parm buf: 문자열 IP를 기록할 메모리 포인터
 * @parm len: @parm buf의 사이즈
 * @parm ip: 정수형 IP
 * @return: None
 */
void
get_ip(char *buf, size_t len, const unsigned char *ip)
{
    snprintf(buf, len, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
}
